/*
   Armadillo fork module
   1999-2000, Randall Maas (randym@acm.org)
*/

#include "dillib.h"
#include <signal.h>

/* @txh @txhcat=Processes */
pid_t __Dillo_fork(const int* ChildIO, unsigned Flags)
{
   /*
     Description
      Basically a UNIX-like fork.  You probably want to use
      @code{__Dillo_fork2()} (@pxref{__Dillo_fork2}) or @code{__Dillo_spawn()}
      (@pxref{__Dillo_spawn}) instead of this routine.

      The caller may pass a pipes which will be used to communicate with
      the child process.  -1 is used to indicate that the pipe's IO channel
      should not be employed.  The pipe-ends are closed after forking.  On fork
      failure all of the pipe ends are closed.

      @var{ChildIO} is used to send and receive data with the child process;
      @var{ChildIO[0]} is read by the child, @var{ChildIO[1]} is written to
      by the child.  @emph{Warning: this is not same as a what you would get
      from @code{pipe()}, but rather from two calls to @code{pipe()}.}
      @var{ChildIO} is dup'd to @file{stdin}, @file{stdout}, @file{stderr},
      and then closed.

      The child process has its @code{SIGCHLD} and @code{SIGPIPE} signals
      set to exit the child process.
      
     Return Value
      -1 on error, 0 if this is the child process, otherwise the process ID
      of the child if we are the parent.

     See Also
      @code{__Dillo_fork2()} (@pxref{__Dillo_fork2}),
      @code{__Dillo_spawn()} (@pxref{__Dillo_spawn})

     Cavaets
      Profiling reveals that this is intensely slow.  Don't fork unless you
      need to.
  */

  pid_t child_pid = fork ();

  /* Return error */
  if (child_pid < 0)
    {
       /* Close the file descriptors since we won't be using them */
       if (ChildIO)
	 {
	   close(ChildIO[0]);
	   close(ChildIO[1]);
	 }

       Err_cmdline(DILLO_FORK_FAILED, strerror(errno));
       return -1;
    }
  
  if (child_pid)
    {
       /* We are the parent */

       /* Close the pipe ends that only the child will use: */ 
       if (ChildIO)
	 {
 	    /* Childs STDOUT */
	    close(ChildIO[0]);
	    close(ChildIO[1]);
         }
       return child_pid;
    }

   /* Redirect the input pipe to the childs stdin */
   if ((Flags &DILLO_NEWIN) &&
       dup2 (ChildIO[0], STDIN_FILENO) != STDIN_FILENO)
     {
        Err_cmdline(DILLO_DUP2_FAILED, "stdin", strerror(errno));
     }

   /* redirect stdout to the pipe we just created */
   if (ChildIO[1] != -1 &&
       dup2 (ChildIO[1], STDOUT_FILENO) != STDOUT_FILENO)
     {
        Err_cmdline(DILLO_DUP2_FAILED, "stdout", strerror(errno));
     }

   /* redirect stderr to the pipe we just created */
   if ((Flags & DILLO_SHARE_ERR) &&
       dup2 (ChildIO[1], STDERR_FILENO) != STDERR_FILENO)
     {
        Err_cmdline(DILLO_DUP2_FAILED, "stderr", strerror(errno));
    }

   /* We are the child... set ourselves up to die cleanly in case of
      badness */
   signal(SIGPIPE, _exit);
   signal(SIGCHLD, _exit);

   /* Note these can be to the /same/ file descriptor */
   close(ChildIO[0]);  /* pipe no longer needed */
   close(ChildIO[1]);
   return 0;
}

