/*
   Armadillo fork module
   1999-2000, Randall Maas

   I'll quote an article that explains things better than I can:
   "Testing UNIX Daemons"
   Len DiMaggio, Dr Dobbs Journal March 2000, p38-44
*/


#include "dillib.h"
#include <sys/ioctl.h>
#include <termios.h>
#include "IOQ.h"

#if defined(TIOCNOTTY) && !defined(HAVE_SETSID)
static void __TTY_detach(void)
{
  /*
     This is used to detach from the controlling tty /only/ if the system
     doesn't have setsid()
  */

  int tty, Ret=0;
  if ((tty = open("/dev/tty", O_RDWR)) == -1) Ret =-1;
  else
   {
     int R2;
     Ret= ioctl(0, TIOCNOTTY, 0);
     R2 = close(tty);
     if (Ret >= 0) Ret = R2;
   }

  if (Ret == -1)
    Err_cmdline(DILLO_TIOCNOTTY_FAILED, strerror(errno));
}
#endif

static int Open_pipe(int* ParentIO, int* ChildIO, unsigned Flags);
static inline int Close_pipe(const int* P, int R)
{
   close(P[0]);
   close(P[1]);
   return -R;
}

/* @txh @txhcat=Processes */
pid_t __Dillo_fork2(int* FileDes, unsigned Flags)
{
   /*
      Description
       Basically a UNIX-like fork.

       A short note on how the child processes' IO is setup.
       The child process will write to its @code{stdout} IO channel.
       If @var{Flags} as the @code{DILLO_NEWIN} flag set, a new pipe will be
       created; the child process will read from its @code{stdin} IO channel.
       Otherwise, it is not defined here what the child will read from.


       @var{FileDes} is set for the parent to communicate with the child;
       it is not set for the child process.  @var{FileDes[0]} is used to read
       the childs output; it is set to be non-blocking, and close on
       @code{exec()}.

       If the @code{DILLO_NEWIN} flag is set, @var{FileDes[1]}
       can be used to to send data to the child.  Otherwise @var{FileDes[1]}
       is set to a meaningless value.

       If the @code{DILLO_SHARE_ERR} flag is set, the child's standard error
       is dup'd to its standard output.  Otherwise, its standard error
       channel is the same as its parent.

       If the @code{DILLO_TTY} flag is set, the child will be set to
       communicate using a master/slave pseudo-tty instead of standard pipe.

       The child process has its @code{SIGCHLD} and @code{SIGPIPE} signals
       set to exit the child process.
      
      Return Value
       -1 on error, 0 if this is the child process, otherwise the process ID
       of the child if we are the parent.

      See Also
       @code{__Dillo_fork()}  (@pxref{__Dillo_fork}),
       @code{__Dillo_spawn()} (@pxref{__Dillo_spawn})

      Cavaets
       Profiling reveals that this is intensely slow.  Don't fork unless you
       need to.

       The child process may choose to fork again.  This is done one linux
       systems when the @code{DILLO_TTY} flag is used, as a part of the
       mechanism to change controlling tty's.

      @end txh
   */

   int ChildIO[2];
   pid_t child_pid;
   struct termios tt;

   if (Open_pipe(FileDes, ChildIO, Flags)) return -1;

   /* Capture the current tty state so we can dup it */
   if ((Flags & DILLO_TTY) && tcgetattr(STDIN_FILENO, &tt) == -1)
     {
        Err_cmdline(DILLO_TCGETATTR_FAILED, strerror(errno));
	return -1;
     }

   if ((child_pid = __Dillo_fork(ChildIO, Flags))
       < 0)
     {
        /* The fork procedure failed, and it spit out an error message,
	   so we will return to the caller.
	*/ 
        return Close_pipe(FileDes, -1);
     }

   if (child_pid)
     {
        /* We are the parent. */
	/* Set the results pipe to be nonblocking & close-on-exec */
	FD_bkgnd(FileDes[0]);

	return child_pid;
     }
 
   /* this is the child */

   /* Close the parents stuff */
   Close_pipe(FileDes,0);

   if ((Flags & DILLO_TTY))
     {
        /* Our pipes are TTY based.  Set original tty modes on the slave tty,
	   so it looks like a real terminal */

        if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &tt) == -1)
	  Err_cmdline(DILLO_TCSETATTR_FAILED, strerror(errno));

	/* Detach the controlling terminal
	   "The process that call[s] setsid is made the session leader of [a]
	    new session and the process group leaer, and it does not have a
	    controlling terminal.  If the process did have a controlling
	    terminal, then calling setsid breaks the connection between the
	    process and the controlling terminal." (DiMaggio)
	 */
        #ifdef HAVE_SETSID
         setsid();
        #elif defined(TIOCNOTTY)
	 /* Break the controlling terminal the older, harder way */
	__TTY_detach();
        #endif

	#ifdef TIOCSCTTY
        #ifdef __linuX__
	 /* there is a strange linuxism: we need to fork again in order for
	    setting the controlling tty (TIOCSCTTY) to work.*/
	 if (fork() > 0) _exit(0); 
	#endif 
	if (ioctl(STDIN_FILENO, TIOCSCTTY, (char*)1) == -1)
	  Err_cmdline(DILLO_IOCTL_FAILED, "TIOCSCTTY", strerror(errno));
	#endif
     }
   
   return 0;
}

static int Open_pipe(int* ParentIO, int* ChildIO, unsigned Flags)
{
   /*
      This routine just opens the pipe, whether it be traditional or via
      a psuedo tty pair
   */
   int fds[2];

   if (!(Flags & DILLO_TTY))
     {
        int ChildOut[2];
        if (pipe (ChildOut) != 0)
	  {
	     /* The pipe creation procedure failed... alert the user */
	     Err_cmdline(DILLO_PIPE_FAILED, strerror(errno));
	     return -1;
	   }
       
	if (Flags & DILLO_NEWIN)
	  {
	     int ChildIn[2];
	     if (pipe(ChildIn) != 0)
	       {
		  /* The pipe creation procedure failed... alert the user */
		  Err_cmdline(DILLO_PIPE_FAILED, strerror(errno));
		  
		  /* Close off the unused pipes */
		  close(ChildOut[0]);
		  close(ChildOut[1]);
		  return -1;
	       }
	     ParentIO[1] = ChildIn [1];
	     ChildIO [0] = ChildIn [0];
	  }
	else
	  {
	     /* The file descriptor aint used */
	     ParentIO[1] = -1;
	     ChildIO [0] = -1;
	  }

       /* Set up the struct to say how we will communicate with the
	  child */
       ParentIO[0] = ChildOut[0];
       ChildIO [1] = ChildOut[1];
       return 0;
     }

   if (__PTY_pipe(fds))
     {
        Err_cmdline(DILLO_TTY_FAILED, strerror(errno));
        return -1;
     }
   ParentIO[0]=fds[0];
   ParentIO[1]=fds[0];
   ChildIO [1]=fds[1];
   ChildIO [0]=fds[1];
   return 0;
}
