/*
   Derived from MSDOS-specific functions for `ls'.

   Original core Written by Eli Zaretskii <eliz@is.elta.co.il>  
   Modified extensively from ls-msdos (sh-utils) Randy Maas (randym@acm.org) to
   work in a "general" fashion: atleast so that I can pipe it thru curses.
*/

#include <string.h>

/*  Screen write redirector.  We need this to support colorization
    without requiring ANSI.SYS driver (or its work-alike) to be loaded.

    The buffer is written directly to
    the screen while converting the ANSI escape sequences into calls to
    macro functions (possibly curses) which change text attributes.  A non-zero
    value is then returned to the caller to signal that the output has been
    handled.

    Warning: this function relies on the fact that `ls' disables the use
    of TAB characters when colorization is required, and therefore it
    does NOT expand TABs!  It also doesn't check for NULL characters in
    the buffer, and might end the output prematurely if there are NULLs.  */

#include <stdio.h>
#include <stdarg.h>
#include <errno.h>
#include <stdlib.h>
#include "platform.h"
#include "vtcurses.h"
//#define FAST_FAIL

#define ESC '\033'

static size_t  _screen_esc  (void* SPtr, const char* Buf, size_t buflen);
static size_t  _screen_escb (void* SPtr, const char* Buf, size_t buflen);
static ssize_t _screen_quant(void *SPtr, const char* Buf, size_t Blen);
static void    _screen_attr (void *SPtr, int code, int *fc, int *bc);
static void set100mode(void* SPtr, int mode);
static void clear100mode(void* SPtr, int mode);


/* @txh */
ssize_t _screen_write (void* SPtr, const char* anchor, size_t Size)
{
  /*
     Description
      Screen write redirector.  We user this to support almost all of the VT100
      terminal codes, including colorization.   The ANSI escape sequences into
      calls to macro functions (possibly curses) which change text attributes.
      A non-zero value is then returned to the caller to signal that the output
      has been handled.

      The buffer pointed to by @var{anchor} must be zero terminated.

     Return Value
      -1 on error,
      >= 0 is the number of bytes processed.
     @end txh
  */

  const char *p_next=anchor;
  size_t buflen=Size;
  WINDOW* Win=(WINDOW*)SPtr;

  if (!anchor)
    {
      errno = EINVAL;
      return -1;
    }

#define Skip(n) p_next += n; buflen -= n; continue;


  /* Walk the buffer, writing text directly to video RAM,
     changing color attributes when an escape sequence is seen.  */
  //  for (; (p_next = memchr (anchor, ESC, buflen)) != 0; anchor = p_next)
   for (;;anchor=p_next)
   {
      size_t S;
      int y=0,x=0;
      for (S=0; S<buflen &&
	     (*p_next >= ' ' || *p_next == '\b');
	   p_next++,S++);

      if (S>=buflen)
	{
	   /* Output what's left in the buffer.  */
	   waddnstr(SPtr, anchor, buflen);
	   return Size;
	}

      if (*p_next == '\r')
	{
	   /* Preserve out current line position so we can return to it later
	    */
	   getyx   (Win,y,x);
	   scrollok(SPtr,0);
	}

      /* If some chars seen since the last escape sequence,
	 write it out to the screen using current text attributes.  */
      if (S)
	{
	   waddnstr(SPtr, anchor, S);
	   buflen -= S;
	}

      switch(*p_next)
      {
         case ESC:
	   if (buflen < 3)
	     return Size-buflen;
      
	   if (!(S =_screen_esc (SPtr, p_next+1,buflen)))
	     return Size -buflen; //We reached the end of the buffer in a code
	   Skip(S);

         case 7:
	   /* Parse beeps */
	   beep();
	   Skip(1);

         case '\t':
	   /* We implement a tab as a skip;
	      This is especially important in "reverse video" modes.  As
	      counter-intuitive as it may sound, the tabs aren't supposed to
	      modified the screen contents (!!)
	   */
	   getyx(Win,y,x);
	   wmove(Win,y,(x+8)-(x+8)%8);
	   Skip(1);

         case '\r':
	   /* Return to the start of the line */
	   wmove    (Win,y,0);
	   scrollok (Win,1);
	   Skip(1);

	 /*
	    In new line mode
	      LF FF VT move the cursor to the first column of the next line
	    In line feed mode:
	      LF FF VT move the cursor to the current column of the next line
	 */
         case '\n': case '\f': case '\v':
	  {
	     int by,bx;
	     getyx   (Win, y, x);

	     /* Check to see if we should scroll */
	     getmaxyx(Win,by,bx);
	     if (y+1>=by)
	       {
		  scrollok(Win,1);
		  scroll  (Win);
		  y--;
	       }
	     if (LineMode == NEW_LINE)
	       wmove    (Win,y+1,0);
	     else
	       wmove    (Win,y+1,x);
	  }

         default:
	   /* Skip junk characters */
	   Skip(1);
      }
    }
#undef Skip
}

/*
  @txh
  @node ANSI escape sequences, screen
  @heading Supported ANSI escape sequences

  @emph{These codes may be incorrect implementations of the standard!
  Please let me know if they are wrong.}  Some of the sequence semantics were
  gleened from the Screen documentation (an excellent program).

  @multitable @columnfractions 0.3 0.6
  @item ESC c	@tab Reset to initial state 
  @item ESC g   @tab @emph{ignored} Visual bell
  @item ESC n   @tab @emph{ignored} Lock shift G2
  @item ESC o   @tab @emph{ignored} Lock shift G3
  @item ESC A	@tab Up one line
  @item ESC k	@tab @emph{ignored} Title definition string
  @item ESC A   @tab Up one line
  @item ESC B	@tab Down one line
  @item ESC C	@tab Right one column
  @item ESC D	@tab Down one line, scrolling if neccessary
  @item ESC E   @tab Next line.
       This moves the cursor to the start of the next line.  If the next line
       is off the display, the screen will be scrolled.  Otherwise the screen
       contents are not modified.

  @item ESC F   @tab @emph{ignored}
  @item ESC G   @tab @emph{ignored}
  @item ESC H   @tab Goto home location
  @item ESC I   @tab @emph{ignored}
  @item ESC J   @tab Clear to beginning of current line
  @item ESC K   @tab Clear to end of current line
  @item ESC M	@tab Up one line, scrolling if neccessary
  @item ESC N   @tab @emph{ignored} Single shift G2
  @item ESC O   @tab @emph{ignored} Single shift G3
  @item ESC P   @tab @emph{ignored} Device control string (string is not interpreted) 
  @item ESC Z   @tab @emph{ignored} 
  @item ESC 1   @tab @emph{ignored}
  @item ESC 2   @tab @emph{ignored}
  @item ESC 7	@tab Save current cursor state
  @item ESC 7p  @tab Make cursor invisible
  @item ESC 8	@tab Restore cursor state
  @item ESC 8p  @tab Make cursor visible
  @item ESC =   @tab @emph{ignored} Application keypad mode
  @item ESC <   @tab @emph{ignored}
  @item ESC >   @tab @emph{ignored} Numeric keypad mode
  @item ESC \   @tab @emph{ignored} String terminator 
  @item ESC ^   @tab @emph{ignored} Privacy message string
  @item ESC !   @tab @emph{ignored} Global message string
  @item ESC _   @tab @emph{ignored} Application program command
  @item ESC ]   @tab @emph{ignored} Operating system command
  @item ESC #8	@tab @emph{ignored} Fill screen with E's
  @item ESC ]R	@tab Reset terminal
  @item ESC [@@ @tab Insert blank character after cursor
  @item ESC [@var{P}@@
                @tab Insert character whose ASCII number is @var{P} after the
		current cursor location
  @item ESC [c  @tab @emph{ignored} Send VT100 Identification string 
  @item ESC [@var{p}c @tab
        @item @tab ?0  Cursor normal
	@item @tab ?1  Cursor invisible
	@item @tab ?8  Cursor big
  @item ESC [@var{P}d	@tab Go to @var{P} row
  @item ESC [f  @tab Go to home location
  @item ESC [@var{row};@var{col}f  @tab Go to specified row and column
  @item ESC [g	@tab @emph{ignored} Clear tab at current column
  @item ESC [0g	@tab @emph{ignored} Clear tab at current column
  @item ESC [3g	@tab @emph{ignored} Clear all tabs
  @item ESC [@var{Ps};@var{...};@var{Ps}h
		@tab @emph{ignored} Set mode
	@item @tab ?25  (Linux console) cursor normal (visible)
	@item @tab ?47	@emph{ignored} (XTERM) switch to alternate screen
  @item ESC [@var{Ps};@var{...};@var{Ps}l
		@tab @emph{ignored} reset mode
	@item @tab ?25  (Linux console) cursor hidden (invisible)
	@item @tab ?47	@emph{ignored} (XTERM) switch from alternate screen to main
  @item ESC [6@var{P1} @tab @emph{ignored} Device status report
  @item ESC [4i	@tab @emph{ignored} Stop relay to printer
  @item ESC [5i	@tab @emph{ignored} Start relay to printer
  @item ESC [@var{Ps}m	@tab Set the character attributes
  @item ESC [6n	@tab @emph{ignored} Send cursor position report 
  @item ESC [@var{Bot}r	@tab Set the scroll region from line 0 to @var{Bot}
  @item ESC [@var{Top};@var{Bot}r	@tab Set the scroll region from line
  	@var{Top} to @var{Bot}
  @item ESC [s  @tab Save cursor and attributes
  @item ESC [8;@var{Ph};@var{Pw}t
	@tab @emph{ignored} Resize the window to @var{Ph} lines and
	@var{Pw} columns
  @item ESC [u  @tab Restore cursor and attributes
  @item ESC [x  @tab @emph{ignored} Send terminal parameter report 
  @item ESC [A  @tab Move cursor up one line
  @item ESC [@var{P}A	@tab Move cursor up @var{P} lines
  @item ESC [B		@tab Move cursor down one line
  @item ESC [@var{P}B	@tab Move cursor down @var{P} lines
  @item ESC [C		@tab Right one column
  @item ESC [@var{P}C	@tab Right @var{P} columns
  @item ESC [D		@tab Left one column
  @item ESC [@var{P}D	@tab Left @var{P} columns
  @item ESC [G          @tab Go to left most column
  @item ESC [@var{P}G	@tab Go to @var{P} column
  @item ESC [H  	@tab Go to home location
  @item ESC [@var{row};@var{col}H  @tab Go to specified row and column
  @item ESC [J   @tab Clear to bottom of screen 
  @item ESC [0J  @tab Clear to bottom of screen 
  @item ESC [1J  @tab @emph{ignored} Clear to beginning of screen 
  @item ESC [2J  @tab Clear entire screen 
  @item ESC [K   @tab Clear to end of current line
  @item ESC [0K  @tab Clear to end of current line
  @item ESC [1K  @tab @emph{ignored} Clear backward to begining of current line
  @item ESC [2K  @tab Clear entire line
  @item ESC [L		@tab Insert line
  @item ESC [@var{P}L	@tab Insert @var{P} lines
  @item ESC [M  	@tab Delete line
  @item ESC [@var{P}M  	@tab Delete @var{P} lines
  @item ESC [P  	@tab Backspace
  @item ESC [@var{P}P  	@tab Backspace @var{P} times
  @item ESC [@var{P1};@var{P2}R   @tab @emph{ignored} Character position report
  @item ESC [U		@tab Next page
  @item ESC [@var{P}U	@tab Next @var{P} pages
  @item ESC [V		@tab Previous page
  @item ESC [@var{P}V   @tab Previous @var{P} pages
  @item ESC [X		@tab Erase 1 character
  @item ESC [@var{P}X	@tab Erase @var{P} characters
  @item ESC [Z  @tab @emph{ignored} Move to previous tab stop
  @item ESC [@var{P}Z  @tab @emph{ignored} Move to previous tab stop
  @item ESC [?  @tab Mode control
  @item ESC [>c  @tab @emph{ignored} Send secondary device attributes string
  @end multitable

  @subheading References
   The escape sequence definitions were derived in part from the following
   sources:
   @itemize @bullet
   @item The @file{screen} documentation
   @item @url{http://www.cs.utk.edu/~shuford/terminal/vt100_reference_card.txt}
   @item @file{/etc/termcap}
   @item @samp{infocmp -L}
   @end itemize

  @end txh
*/

 
static size_t _screen_esc (void* SPtr, const char* Buf, size_t buflen)
{
   /* "Esc-[" sequence */
   switch (*Buf)
   {
      case '[':
        if (buflen <= 3)
	  return 0;
      	return _screen_escb(SPtr, Buf+1,buflen);
      case ']': __VT_reset(SPtr);      return 3;
      case 'c': __VT_reset(SPtr);      return 2;

      /* Look into the VT52 compatibility items */
      case 'M':
      case 'A': __VT_dmove(SPtr,-1,0); return 2; /* Up one */
      case 'D':
      case 'B': __VT_dmove(SPtr, 1,0); return 2; /* down one line */
      case 'C': __VT_dmove(SPtr, 0,1); return 2; /* right one */
      case 'E': __VT_nextline(SPtr);   return 2; /* next line */
      case 'H': wmove(SPtr,0,0);       return 2;
      case 'J': wclrtobot(SPtr);       return 2;
      case 'K': __VT_clreol(SPtr);     return 2;
      case '7':
	if (Buf[1] == 'p')
	  {
	     /* Hide the cursor */
	     __VT_chide(SPtr);
	     return 3;
	  }
	/* Save cursor state and attributes */
	__VT_cursave(SPtr);
	return 2;

      case '8':
	if (Buf[1] == 'p')
	  {
	     /* Show the cursor */
	     __VT_cshow(SPtr);
	     return 3;
	  }
	/* Restore cursor state and attributes */
	__VT_currest(SPtr);
	return 2;

      case '>': /* These two are application keypad things that we ignore */
      case '=': return 2;
      case '(':
      case ')':
      case '*':
      case '+':
      case '#': return 3;
      default: /* Especially: E F G I K M Z 1 2 7 8 = > < */
#ifdef FAST_FAIL
	fprintf(stderr, "ESC %c\n", *Buf); exit(-1);
#endif
	return 2;
   }
}

static size_t _screen_escb(void* SPtr, const char* Buf, size_t buflen)
{
   switch (*Buf)
   {
      case '@': __VT_ins (SPtr,1); return 3; /* Add char after */
      case 'c':                    return 3; //send esc[?6c backs
      case 'd': __VT_mrow(SPtr,0); return 3; 
      case 'G': __VT_mcol(SPtr,0); return 3; 
      case 'f':
      case 'H': wmove  (SPtr,0,0); return 3;
      case 'J': wclrtobot  (SPtr); return 3;
      case 'K': __VT_clreol(SPtr); return 3;
      case 'L': winsertln  (SPtr); return 3;
      case 'M': wdeleteln  (SPtr); return 3;
      case 'P': __VT_backspace(SPtr,1);return 3;
      case 'A': __VT_dmove(SPtr,-1,0); return 3; /* Up one */
      case 'B': __VT_dmove(SPtr, 1,0); return 3; /* down one line */
      case 'C': __VT_dmove(SPtr,0, 1); return 3; /* right one */
      case 'D': __VT_dmove(SPtr,0,-1); return 3; /* left one */
      case 'U': __VT_nextpage(SPtr,1); return 3; /* Next Page */
      case 'V': __VT_prevpage(SPtr,1); return 3; /* Previous Page */
      case 'X': wdelch(SPtr);          return 3; /* Erase character */
      case 's': __VT_cursave(SPtr);    return 3; /* Save cursor state */
      case 'u': __VT_currest(SPtr);    return 3; /* Restore cursor state */

	/*
      case 'Z': move to previous tab stop;
	*/
      case '>': /* Code to ignore "send secondary device" string */
	if (buflen < 4)
	  return 0; //We reached the end of the buffer in a code
	return 4;

      case '?': // Mode control
	{
	  ssize_t S;
	  if ((S = _screen_quant(SPtr, Buf+1, buflen-3)) <0)
	    return 0; //We reached the end of the buffer in a code
	  return S+3; /* Count the size of the ^[?...m */
	}
	/* See if it is a quantified movement, coloration, etc.
	   For some of these, we check for the sequences that don't need to be
	   converted to integers
	*/
      case '0':
        if (buflen < 4)
	  return 0; //We reached the end of the buffer in a code
	switch (Buf[1])
	{
	   case 'g':                    return 4;
	   case 'J': wclrtobot  (SPtr); return 4;
	   case 'K': __VT_clreol(SPtr); return 4;
	}
	goto quant;
      case '1':
        if (buflen < 4)
	  return 0; //We reached the end of the buffer in a code
	if (Buf[1] == 'J') {
#ifdef FAST_FAIL
	  fprintf(stderr, "Clear to beginning of screen\n"); exit(-1);
#endif
	  return 4;}
	else if (Buf[1] == 'K') {
#ifdef FAST_FAIL
	       fprintf(stderr, "Clear to beginning of line\n"); exit(-1);
#endif
	  return 4;}
	goto quant;
      case '2':
        if (buflen < 4)
	  return 0; //We reached the end of the buffer in a code
	     if (Buf[1] == 'J') {werase(SPtr);         return 4;}
	else if (Buf[1] == 'K') {__VT_eraseline(SPtr); return 4;}
	quant:

      case '3':
      case '4': case '5': case '6':
      case '7': case '8': case '9':
	/* get past "Esc-[" sequence ditto, plus the end marker */
	if (buflen < 4)
	  return 0; //We reached the end of the buffer in a code
      {
	 ssize_t S;
	 if ((S = _screen_quant(SPtr, Buf, buflen-2)) < 0)
	   return 0; //We reached the end of the buffer in a code
	 return S+2; /* Count the size of the ^[...m */
      }

      case 'r': __VT_scrollreset(SPtr); return 3;
      case 'm':
	__VC_cattrset(SPtr, A_NORMAL);
	__VC_color   (SPtr, COLOR_WHITE, COLOR_BLACK);
	return 3;
      default: /* include c n q g */
#ifdef FAST_FAIL
	    write(2, Buf-2,buflen);
	    fprintf(stderr, "\n"); exit(-1);
#endif
        return 3; /* Skip the bogon */
      }
}

static ssize_t _screen_quant(void *SPtr, const char* Buf, size_t Blen)
{
   /*
      Description
       Parse the ANSI codes that involve numbers to quantify or represent
       something.

      Return Value
   */

  const char* p = Buf;
  int fc=-1, bc=-1, Ret=0;
  long last_code=-1;

  while (Blen)
   {
      char *q;
      long code;

      code = strtol (p, &q, 10);
      if (q-p+1 > Blen) return -1;

      /* Sanity checks:
	 
	 q > p unless p doesn't point to a number;
	 ANSI codes are between 0 and 47;
	 Each ANSI color code ends with a `;' or an `m'.
	 The movement ANSI code ends with a 'P'

	 If any of the above is violated, we just ignore the bogon. */
      if (q == p || code < 0)
	{
	  return q - Buf;
	}

      /* See if it ends the escape sequence */
      Ret = q-Buf+1;
      switch (*q)
      {
	case ';':
	  /* more codes to follow */
	  q++;
	  /* Check for overflow */
	  if (q-p > Blen) return -1;
#if 0
	  if (last_code != -1)
	    {
	       /* assume that this will be an attribute modifier */ 
	       if (last_code > 50)
		 return q-Buf;
      	       _screen_attr(SPtr,last_code,&fc, &bc);
	    }
#endif
	  last_code = code;
	  break;

        case 'g': return Ret;
	case 'h':
	   if (last_code != -1) set100mode(SPtr, last_code);
	   set100mode(SPtr, code);
	   return Ret;
	case 'l':
	   if (last_code != -1) clear100mode(SPtr, last_code);
	   clear100mode(SPtr, code);
	   return Ret;
	  
	case 'm':
	  /* ANSI color codes are between 0 and 49 */
	  if (last_code >=0 && last_code < 49)
	    {
	       _screen_attr(SPtr,last_code,&fc, &bc);
	       last_code = -1;
	    }
	  if (code > 49)
	    return q - Buf;
	  _screen_attr(SPtr,code,&fc,&bc);
	  __VC_color(SPtr, fc,bc);
	  return Ret;

        case '@': /* Add char after */
	  __VT_ins(SPtr,code);
	  return Ret;

        case 'A': __VT_dmove(SPtr,-code,0); return Ret; /* Move Up */
        case 'B': __VT_dmove(SPtr, code,0); return Ret; /* Move down */
        case 'C': __VT_dmove(SPtr,0, code); return Ret; /* Move cursor right */
        case 'D': __VT_dmove(SPtr,0,-code); return Ret; /* Move cursor left */
        case 'd': __VT_mrow (SPtr,code-1);  return Ret; /* Goto row */ 
        case 'G': __VT_mcol (SPtr,code-1);  return Ret; /* Goto column */ 
        case 'U': __VT_nextpage(SPtr, code);return Ret; /* Next Page */
        case 'V': __VT_prevpage(SPtr, code);return Ret; /* Previous Page */
 
        case 'H': /* Move the cursor */
        case 'f':
	  if (last_code != -1) wmove(SPtr, last_code-1, code-1);
	  return Ret;

        case 'L':
	  while (code--) winsertln(SPtr);
	  return Ret;

        case 'M':
	  while (code--) wdeleteln(SPtr);
	  return Ret;

	case 'P': /* Move left, delete char under cursor */
	  __VT_backspace(SPtr,code);
	  return Ret;

        case 'X':
	  for (;code>0;code--) wdelch(SPtr);
	  return Ret;

        case 'r': /* Define a scroll region */
	  __VT_scrreg(SPtr, last_code==-1?0:last_code-1, code==-1?-1:code-1);
	  return Ret;

        case 'c':
	  switch (code)
	   {
	      case 0: __VT_cshow(SPtr); return Ret;
	      case 1: __VT_chide(SPtr); return Ret;
	      case 8: curs_set(2); return Ret;
	      default: break;
	   }
	  return Ret;

	default:
	  /* Mysterious codes that are ignore 4l == end insert mode */
#ifdef FAST_FAIL
	    write(2,p,Blen);
	    fprintf(stderr, "\n"); exit(-1);
#endif
	  return Ret;
	}

      Blen -= q-p;
      p=q;
    }
  
  return -1;
  //  return p-Buf+1;
}

void _screen_attr(void *SPtr, int code, int *fc, int *bc)
{
   /* Convert ANSI codes to color fore- and background.  */
   /* Only dark colors mentioned here, so that bold has visible effect.  */
   static short screen_color[] = {COLOR_BLACK, COLOR_RED,  COLOR_GREEN,
				  COLOR_YELLOW,COLOR_BLUE, COLOR_MAGENTA,
				  COLOR_CYAN,  COLOR_WHITE};
   switch (code)
   {
     case 0:       /* all attributes off -- including color */
       __VC_cattrset(SPtr, A_NORMAL);
	*fc = COLOR_WHITE;
	*bc = COLOR_BLACK;
       break;
     case 1:       /* intensity on */
       __VC_cattron (SPtr, A_BOLD);
       break;
     case 21:      /* intensity off */
     case 22:
     case 23:
       __VC_cattroff(SPtr, A_BOLD);
       break;
     case 4:       /* underline on */
       __VC_cattron (SPtr, A_UNDERLINE);
       break;
     case 24:       /* underline off */
       __VC_cattroff(SPtr, A_UNDERLINE);
       break;
     case 5:       /* blink */
       __VC_cattron (SPtr, A_BLINK);
       break;
     case 25:       /* blink */
       __VC_cattroff(SPtr, A_BLINK);
       break;
     case 7:       /* reverse video */
       __VC_cattron (SPtr, A_REVERSE);
       break;
     case 27:      /* reverse off */
       __VC_cattroff(SPtr, A_REVERSE);
       break;
     case 8:       /* concealed on */ 
       /* They don't *really* want it invisible, do they?  */
       __VC_cattron (SPtr, A_INVIS);
       break;
     case 30: case 31: case 32: case 33: /* foreground color */
     case 34: case 35: case 36: case 37:
       *fc = screen_color[code-30];
       break;
     case 39: *fc = COLOR_WHITE; break;
     case 40: case 41: case 42: case 43: /* background color */
     case 44: case 45: case 46: case 47:
       *bc = screen_color[code - 40];
       break;
     case 49: *bc = COLOR_BLACK; break;
     default:
       /* ignore unknown codes (10) */
       break;
     }
}


static void set100mode(void* SPtr, int mode)
{
   switch(mode)
    {
       case 25: __VT_cshow(SPtr); break;
       case 20: LineMode=NEW_LINE; break;
       default:
#ifdef FAST_FAIL
	    fprintf(stderr, "100 mode: %d\n", mode); //exit(-1);
#endif
    }
}

static void clear100mode(void* SPtr, int mode)
{
   switch(mode)
    {
       case 25: __VT_chide(SPtr); break;
       case 20: LineMode=LINE_FEED; break;
       default:
#ifdef FAST_FAIL
	    fprintf(stderr, "100 mode: %d\n", mode); //exit(-1);
#endif
    }
}

