/*
   Curses stuff to make a virtual terminal work
   2000, Randall Maas (randym@acm.org)

*/
#define def___VT_dmove
#define def___VT_scrollpage
#include "vtcurses.h"

static __VT_cstate_t Save={1,0},Cur={0};
static int* PUsage=NULL; /* Cache data for color pairs */

/* @txh */
void __VT_init (void *SPtr)
{
   /*
      Description
       This routine is used to set up any neccessary structures for the
       virtual terminal.
   */

   if (!PUsage) PUsage=calloc(COLOR_PAIRS,sizeof(*PUsage));
}

void __VT_reset(void* SPtr)
{
   /*
      Description
	This resets the virtual terminal to its initial settings, including
	the scrolling region, cursor visibility, color palette, and clears the
	screen.
   */
   int x,y;
   getyx((WINDOW*)SPtr,y,x);
   curs_set(Cur.Visibility=1);
   wsetscrreg((WINDOW*)SPtr,Cur.Top=0,Cur.Bottom=y); 
   use_default_colors();
   wattr_set((WINDOW*)SPtr, 0,0,NULL);
   wbkgdset((WINDOW*)SPtr, ' ');
   werase((WINDOW*)SPtr);
}

void __VT_cursave(void* SPtr)
{
   /*
      Description
       This save the current state of the cursor:
       @itemize @bullet
       @item its location,
       @item graphic rendtion:
        @itemize @bullet
         @item foreground and background colors,
	 @item emphasis attributes (bold, italic, etc).
	 @item the cursor visibility and size
        @end itemize
       @item which character set(s) to employ
       @item origin mode flag
       @item wrap flag
       @item state of selective erase
       @end itemize
   */

   Save=Cur;
   wattr_get((WINDOW*)SPtr, &Save.Attr, &Save.Pair, NULL);
   getyx    ((WINDOW*)SPtr, Save.y, Save.x);
}

void __VT_currest(void* SPtr)
{
   /*
      Description
       This restores the previous state of the cursor:
       @itemize @bullet
       @item its location,
       @item graphic rendtion:
        @itemize @bullet
         @item foreground and background colors,
	 @item emphasis attributes (bold, italic, etc).
	 @item the cursor visibility and size
        @end itemize
       @item which character set(s) to employ
       @item origin mode flag
       @item wrap flag
       @item state of selective erase
       @end itemize

	(Saves and restores can not be nested at this time)
   */

   curs_set(Save.Visibility);
   wattr_set((WINDOW*)SPtr, Save.Attr, Save.Pair, NULL);
   wmove    ((WINDOW*)SPtr, Save.y, Save.x);
   Cur=Save;
   wsetscrreg((WINDOW*)SPtr, Cur.Top,Cur.Bottom);
}

void __VT_chide(void* SPtr)
{
   /*
      Description
       Hides the cursor

      See Also
       @code{__VT_cshow()} (@pxref{__VT_cshow})
   */
   curs_set(Cur.Visibility=0);
}

void __VT_cshow(void* SPtr)
{
   /*
      Description
       Makes the cursor visible again

      See Also
       @code{__VT_chide()} (@pxref{__VT_chide})
   */
   curs_set(Cur.Visibility=1);
}

void __VT_scrreg(void* SPtr, int top, int bot)
{
   /*
      Description
       Set the scrolling region for the window.  If @var{bot} is -1, then the
       bottom will be set to the bottom of the window.
   */

   if (bot < 0)
     {
        int Rows, Cols;
	getmaxyx((WINDOW*)SPtr,Rows,Cols);
	bot=Rows-1;
     }
   Cur.Top   =top;
   Cur.Bottom=bot;
   wsetscrreg((WINDOW*)SPtr,top,bot);
}

void __VT_scrollreset(void* SPtr)
{
   /*
      Description
       Resets the scrolling region.  The scrolling region will be set from the
       top of the display window to the bottom of the display window.
   */
   int x;
   Cur.Top=0;
   getmaxyx((WINDOW*)SPtr,Cur.Bottom,x);
   wsetscrreg((WINDOW*)SPtr,Cur.Top,Cur.Bottom);
}

void __VT_clreol(void *SPtr)
{
   /*
      Description
       This clears all of the characters from the current cursor location to
       the end of the line.

      See Also
       @code{__VT_backspace()} (@pxref{__VT_backspace})
   */

   int y, x;
   getyx    ((WINDOW*)SPtr, y,x);
   wclrtoeol(SPtr);
}


void __VT_backspace(void* SPtr, unsigned n)
{
   /*
      Description
       This deletes @var{n} characters underneath the cursor and to the left
       of the cursor (towards the beginning of the line).

       For example, if @var{n} is 1, it will delete the character under the
       cursor and move to the left by one character.

       It does not wrap around.

      See Also
       @code{__VT_clreol()} (@pxref{__VT_clreol})
   */
   int x,y,i=n;
   WINDOW* WPtr = (WINDOW*) SPtr;

   getyx(WPtr,y,x);
   x=(x+1)>i?(x+1)-i:0;
   while(i--) wdelch(WPtr);
   wmove(WPtr,y,x);
}

void __VT_ins(void* SPtr,unsigned n)
{
   /*
      Description
       Insert @var{n} spaces to the right of the cursor
   */
   int i=n,y,x;
   WINDOW* WPtr = (WINDOW*) SPtr;
   getyx(WPtr,y,x);
   while(i--) winsch(WPtr,' '); 
   wmove(WPtr,y,x);
}

int LineMode=LINE_FEED;
static int __VC_color_alloc(short fc, short bc);

int __VC_color(void* SPtr, short fc, short bc)
{
   /*
      Description
       Allocates a color pair and sets the window's active pointer to this

      Return Value
       The color pair

      Cavaets
       The color allocation scheme is a least-recently-used scheme.  If you
       use too many color pairings on the screen, some will eventually become
       ``aliased'' to the same pairing.  That means some colors will get hosed.
      @end txh
   */
   int use_pair=-1,
	cpair;	/* Current color pair being used */
   attr_t a;	/* Current attribute of the window */
   WINDOW* Win = (WINDOW*) SPtr;
   chtype c = ' '; /* Background character */
   static int CallCount=0;
   static short CBC=-1,CFC=-1;

   if (fc == -1 && bc == -1) return 0;

   /* Find out the windows current attributes -- colors and background
	character */
   if (Win)
     wattr_get(Win, &a, &cpair, NULL);
    else
     attr_get(      &a, &cpair, NULL);

   /* Determine the current background */
   if (Win) c =getbkgd(Win);

   /* Remove the colors from the attribute */
   if (!cpair) cpair = PAIR_NUMBER(c);
   a &= ~ COLOR_PAIR(cpair);
   c &= ~ COLOR_PAIR(cpair);

   if ((fc == -1 || bc == -1) && (CFC==-1 || CBC==-1))
     {
	/* If either the foreground or the background color should remain the
	   same, we need to fill it in with the current setting */
	pair_content(cpair, &CFC, &CBC);
     }
   if (fc == -1) fc = CFC; else CFC=fc;
   if (bc == -1) bc = CBC; else CBC=bc;

   use_pair = __VC_color_alloc(fc,bc);
   if (Win)
     {
        /* Determine the background character */
	wbkgdset(Win, c|COLOR_PAIR(use_pair));
        wattron (Win, a|COLOR_PAIR(use_pair));
     }
   else
     {
	bkgdset(c|COLOR_PAIR(use_pair));
        attron (a|COLOR_PAIR(use_pair));
     }

   /* Update our usage count for the color pair */
   PUsage[use_pair]=++CallCount;
   return use_pair;
}

static int __VC_color_alloc(short fc, short bc)
{
   int i, use_pair;
   short F,B;

   /* See if this pair already exists */
   for (i = 0; i < COLOR_PAIRS && (pair_content(i, &F,&B) == OK); i++)
     if (F == fc && B == bc)
       return i;

   /* Allocate a new color -- not very trust worthy.
      Must start with 1.*/
   use_pair=1;
   for (i = 2; i < COLOR_PAIRS; i++)
     {
       if (PUsage[i] < PUsage[use_pair]) use_pair=i; 
     } 
   
   init_pair(use_pair, fc, bc);
   return use_pair;
}
