/*
   Circular Queue for requeusted tasks
   1998-2000 Randall Maas (randym@acm.org)

   The buffer is something like:

   +--------------------------------+
   |                                |
   +--------------------------------+
   ^   ^Head1      ^Head2    ^Tail  ^
   |   (First      (After           |
   |    Actively    last one        |
   |    being       activel         |
   |    worked on)  being           |
   |                worked on)      |
   |                                |
   +--Begin                         +- End

   Stuff is placed at the tail of the queue.  This designed allows a single
   submitter with a single full-fillment and helps ensure that the submitted
   tem is never lost.  The submitter should only use the tail.  The processor
   (not the CPU) can rearrange the items between Head1 and Head2

*/

#ifndef __CircQueue_h
# define __CircQueue_h

/* The simplest circular queue definition.  Like the one below, the head is
   last.  This is specifically to allow single head routines (like IO) and
   multi-head routines to interwork
*/
typedef struct
{
  char* Begin, *End,*Tail, *Head;
} __CQ_t;

#define ___CQ_write_size(Ptr, Head) (Head> (Ptr)->Tail?\
			      (Ptr)->End - Head:\
			      (Ptr)->Tail- Head)

#define __CQ_write_size(Ptr) ___CQ_write_size(Ptr,(Ptr)->Head)


#ifndef def___CQ_read
EXTERN
#endif
ssize_t __CQ_read(int FD, __CQ_t* Ptr)
{
   /*
      Description
       Reads from the specified IO channel, @var{FD}, and places the data
       into the specified circular queue.

      Return Value
       -1 on error, otherwise the number of bytes read into the circular
       buffer.
    */

   ssize_t ToRead;
   if (Ptr->Tail >= Ptr->Head)
     {
        ToRead = Ptr->End-Ptr->Tail;
	if (!ToRead) return 0;
	if (Ptr->Head == Ptr->Begin) ToRead--;
     }
   else
     if ((ToRead = (Ptr->Head-Ptr->Tail)-1) == 0) return 0;
   
   ToRead = read(FD, Ptr->Tail, ToRead);
   if (ToRead <= 0) return ToRead;

   Ptr->Tail += ToRead;
   if (Ptr->Tail == Ptr->End) Ptr->Tail = Ptr->Begin;
   return ToRead;
}

#ifndef def___CQ_write
EXTERN
#endif
ssize_t __CQ_write(int FD, __CQ_t* Ptr)
{
   /*
      Description
       This writes from the circular queue (@var{Ptr}) to the specified IO
       channel (@var{FD}).

      Return Value
       -1 on error, otherwise the number of bytes written.
    */

   ssize_t ToWrite = write(FD, Ptr->Head, __CQ_write_size(Ptr));
   if (ToWrite <= 0) return ToWrite;

   Ptr->Head += ToWrite;
   if (Ptr->Head == Ptr->End) Ptr->Head = Ptr->Begin;
   return ToWrite;
}


# define Define_Queue(Name,NHeads,ElemType) \
typedef struct \
{\
   ElemType* Begin, /* Pointer to the start of the buffer area */\
           * End,   /* Pointer to the end of the buffer area */  \
           * Tail,  /* Pointer to the first place to submit */   \
           * Head[NHeads]; /* Pointers to the items in process */\
} Name;


/*
   Description
    This implements the equivalent of the traditional C command
      *P++
    but for circular queues, which must wrap around.
*/

# define CQueue_append(X,V) \
        ((X)->Tail == (X)->Head[0]?NULL: \
         (*(X)->Tail=V,                  \
          (X)->Tail+1!=(X)->End?(X)->Tail++:((X)->Tail=(X)->Begin,(X)->End))\
         ))
/*
   Description
    This appends an item onto the end of the circular queue.  It is
    implemented as a macro.

   Return Value
    NULL on error, otherwise a pointer the queue slot where the value was
    appended.
*/
   

# define CQueue_first(X) (*(X)->Head[0])
/*
   Description
    This returns the first uncompleted item in the queue.

   Return Value
    The value stored in the first circular queue spot
 */

# define CQueue_next(X) (AryLen((X)->Head)==1)?\
                         ((X)->Head[0]  ==(X)->Tail?NULL:\
                          (X)->Head[0]+1!=(X)->End ? (X)->Head[0]++:\
                          ((X)->Head[0]=(X)->Begin,(X)->End)):\
                         ((X)->Head[1]  ==(X)->Tail?NULL:\
                          (X)->Head[1]+1!=(X)->End ? (X)->Head[1]++:\
                          ((X)->Head[1]=(X)->Begin,(X)->End)))
/*
   Description
    This returns a pointer to the next uncompleted item in the queue.  If
    there is only one head pointer, this will be incremented, effectively
    dequeuing the item.  If there is a second head pointer, it will increment
    that; the item will be effectively marked as `in process' but not dequeued.

  Return Value
   NULL if there is no more items, otherwise a pointer to the next item in
   the queue.
*/

#endif
