/*
   Module to manage the display of the curses window/panel to a shell
   Copright (c) 2000 Randall Maas (randym@acm.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


   Portions -- the change_window code -- are from Fraser McCrossan's work.
*/
#include <sys/ioctl.h>
#include <signal.h>
#include "psh.h"

static int ShellFD=-1;
static WINDOW* ShellWin=NULL;
PANEL* ShellPanel=NULL;

void __ShellWin_exit(void)
{
   /* Close the shell IO channel */
   if (ShellFD != -1)
     {
       close(ShellFD);
       ShellFD = -1;
     }
   del_panel(ShellPanel);
   delwin(ShellWin);
}

void __ShellWin_init(int fd)
{
   ShellFD = fd;
   ShellWin= newwin(LINES-3,COLS,0,0);
   ShellPanel=new_panel(ShellWin);
   scrollok(ShellWin,TRUE);

   change_window(SIGWINCH);
}

/* @txh */
void Shell_print(__CQ_t* Ptr)
{
   /*
      Description
       This prints the output from the shell to our display, including
       interpretation of the VT100 terminal codes.  @var{Ptr} points
       to the circular queue containing the data from the shell subprocess.
       This is not guaranteed to emit all of the data in the queue, since
       portions may require more data to fully interpret the VT100 commands.
      @end txh
   */

   ssize_t S;

   /* Zero terminate the buffer */
   Ptr->Head[__CQ_write_size(Ptr)]=0;

   if ((S= _screen_write(ShellWin, Ptr->Head, __CQ_write_size(Ptr))) < 0)
     return;

   Ptr->Head += S;
   if (Ptr->Head == Ptr->End)
     {
        Ptr->Head = Ptr->Begin;
	if (Ptr->Head != Ptr->Tail)
	  Shell_print(Ptr);
     }
	
   if (Ptr->Tail < Ptr->Head && Ptr->Tail > Ptr->Begin)
     {
        /* There is a bug condition where the buffer wraps around; the virtual
	   screen won't know what to do.  We do: just copy and write
        */
        size_t Asize=__CQ_write_size(Ptr),Bsize= Ptr->Tail-Ptr->Begin;
        char*  Buf = alloca(Asize+Bsize+2);
	memcpy(Buf, Ptr->Head, Asize);
	memcpy(Buf+Asize, Ptr->Begin,Bsize);
	
	/*Zero terminate the buffer */
	Buf[Asize+Bsize]=0;
	if ((S = _screen_write(ShellWin, Buf, Asize+Bsize)) <= 0) return;

	Ptr->Head += S;
	if (Ptr->Head>=Ptr->End)
	  Ptr->Head = Ptr->Begin + (Ptr->Head-Ptr->End);
     }
}

void change_window(int sig)
{
    int pgrp;

    /* First update our child shell's window size */
    struct winsize tty_winsize;
    (void) ioctl(STDIN_FILENO, TIOCGWINSZ, &tty_winsize);
    /* The child's window is 3 smaller so we need to tell it that. */
    tty_winsize.ws_row-= 3;
    (void) ioctl(ShellFD, TIOCSWINSZ, &tty_winsize);

    /* Next signal the child shell (ring its doorbell).  We do this to the
     whole process group */
    (void) ioctl (ShellFD, TIOCGPGRP, (char *) &pgrp);
    (void) killpg(pgrp, SIGWINCH);
    signal(SIGWINCH, change_window);
}
